#!/usr/bin/env python
# coding=utf-8
# Copyright 2023 The HuggingFace Inc. team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and

import argparse
import logging
import math
import os
import random
from pathlib import Path

import accelerate
import datasets
import numpy as np
import torch
import torch.nn.functional as F
import torch.utils.checkpoint
import transformers
from accelerate import Accelerator
from accelerate.logging import get_logger
from accelerate.utils import ProjectConfiguration, set_seed
from datasets import load_dataset
from huggingface_hub import create_repo, upload_folder
from packaging import version
from torchvision import transforms
from tqdm.auto import tqdm
from transformers import T5EncoderModel, T5Tokenizer
from diffusers.utils import pt_to_pil
import diffusers
from diffusers.loaders import AttnProcsLayers
from diffusers import AutoencoderKL, DDPMScheduler, DiffusionPipeline, UNet2DConditionModel
from diffusers.optimization import get_scheduler
from diffusers.training_utils import EMAModel
from diffusers.utils import check_min_version, deprecate, is_wandb_available
from diffusers.utils.import_utils import is_xformers_available
from torch_utils.ambient_diffusion import get_random_mask, get_patch_mask
from diffusers.models.attention_processor import LoRAAttnProcessor
import dnnlib
from training.dataset import ImageFolderDataset
from dnnlib.util import save_image, save_images


if is_wandb_available():
    import wandb



# Will error if the minimal version of diffusers is not installed. Remove at your own risks.
check_min_version("0.17.0.dev0")

logger = get_logger(__name__, log_level="INFO")

DATASET_NAME_MAPPING = {
    "lambdalabs/pokemon-blip-captions": ("image", "text"),
}


def log_validation(pipeline, args, accelerator, epoch):
    unet_type = pipeline.unet.dtype
    pipeline.unet = pipeline.unet.to(torch.float16)
    if args.seed is None:
        generator = None
    else:
        generator = torch.Generator(device=accelerator.device).manual_seed(args.seed)

    logger.info("Running validation... ")

    
    images = []
    for i in range(len(args.validation_prompts)):
        image = pipeline(args.validation_prompts[i], output_type="pt", generator=generator, num_inference_steps=20).images
        images.append(image[0])
        # pil_image[0].save(f"test_{i}.png")

    for tracker in accelerator.trackers:
        if tracker.name == "tensorboard":
            np_images = np.stack([np.asarray(img) for img in images])
            tracker.writer.add_images("validation", np_images, epoch, dataformats="NHWC")
        elif tracker.name == "wandb":
            tracker.log(
                {
                    "validation": [
                        wandb.Image(image, caption=f"{i}: {args.validation_prompts[i]}")
                        for i, image in enumerate(images)
                    ]
                }
            )
        else:
            logger.warn(f"image logging not implemented for {tracker.name}")
    pipeline.unet = pipeline.unet.to(unet_type)
    torch.cuda.empty_cache()


def parse_args():
    parser = argparse.ArgumentParser(description="Simple example of a training script.")
    parser.add_argument(
        "--dataset_name",
        type=str,
        default=None,
        help=(
            "The name of the Dataset (from the HuggingFace hub) to train on (could be your own, possibly private,"
            " dataset). It can also be a path pointing to a local copy of a dataset in your filesystem,"
            " or to a folder containing files that 🤗 Datasets can understand."
        ),
    )
    parser.add_argument(
        "--dataset_config_name",
        type=str,
        default=None,
        help="The config of the Dataset, leave as None if there's only one config.",
    )
    parser.add_argument(
        "--train_data_dir",
        type=str,
        default=None,
        help=(
            "A folder containing the training data. Folder contents must follow the structure described in"
            " https://huggingface.co/docs/datasets/image_dataset#imagefolder. In particular, a `metadata.jsonl` file"
            " must exist to provide the captions for the images. Ignored if `dataset_name` is specified."
        ),
    )
    parser.add_argument(
        "--image_column", type=str, default="image", help="The column of the dataset containing an image."
    )
    parser.add_argument(
        "--caption_column",
        type=str,
        default="text",
        help="The column of the dataset containing a caption or a list of captions.",
    )
    parser.add_argument(
        "--empty_text",
        default=False,
        action="store_true",
        help=(
            "If true, it finetunes the model with empty strings as captions."
        ),
    )
    parser.add_argument(
        "--max_train_samples",
        type=int,
        default=None,
        help=(
            "For debugging purposes or quicker training, truncate the number of training examples to this "
            "value if set."
        ),
    )
    parser.add_argument(
        "--validation_prompts",
        type=str,
        default=None,
        help=("A set of prompts evaluated every `--validation_epochs` and logged to `--report_to`."),
    )
    parser.add_argument(
        "--output_dir",
        type=str,
        default="sd-model-finetuned",
        help="The output directory where the model predictions and checkpoints will be written.",
    )
    parser.add_argument(
        "--cache_dir",
        type=str,
        default=None,
        help="The directory where the downloaded models and datasets will be stored.",
    )
    parser.add_argument("--seed", type=int, default=None, help="A seed for reproducible training.")
    parser.add_argument(
        "--resolution",
        type=int,
        default=512,
        help=(
            "The resolution for input images, all the images in the train/validation dataset will be resized to this"
            " resolution"
        ),
    )
    parser.add_argument(
        "--center_crop",
        default=False,
        action="store_true",
        help=(
            "Whether to center crop the input images to the resolution. If not set, the images will be randomly"
            " cropped. The images will be resized to the resolution first before cropping."
        ),
    )
    parser.add_argument(
        "--random_flip",
        action="store_true",
        help="whether to randomly flip images horizontally",
    )
    parser.add_argument(
        "--train_batch_size", type=int, default=16, help="Batch size (per device) for the training dataloader."
    )
    parser.add_argument("--num_train_epochs", type=int, default=100)
    parser.add_argument(
        "--max_train_steps",
        type=int,
        default=None,
        help="Total number of training steps to perform.  If provided, overrides num_train_epochs.",
    )
    parser.add_argument(
        "--gradient_accumulation_steps",
        type=int,
        default=1,
        help="Number of updates steps to accumulate before performing a backward/update pass.",
    )
    parser.add_argument(
        "--gradient_checkpointing",
        action="store_true",
        help="Whether or not to use gradient checkpointing to save memory at the expense of slower backward pass.",
    )
    parser.add_argument(
        "--learning_rate",
        type=float,
        default=1e-4,
        help="Initial learning rate (after the potential warmup period) to use.",
    )
    parser.add_argument(
        "--scale_lr",
        action="store_true",
        default=False,
        help="Scale the learning rate by the number of GPUs, gradient accumulation steps, and batch size.",
    )
    parser.add_argument(
        "--lr_scheduler",
        type=str,
        default="constant",
        help=(
            'The scheduler type to use. Choose between ["linear", "cosine", "cosine_with_restarts", "polynomial",'
            ' "constant", "constant_with_warmup"]'
        ),
    )
    parser.add_argument(
        "--lr_warmup_steps", type=int, default=500, help="Number of steps for the warmup in the lr scheduler."
    )
    parser.add_argument(
        "--snr_gamma",
        type=float,
        default=None,
        help="SNR weighting gamma to be used if rebalancing the loss. Recommended value is 5.0. "
        "More details here: https://arxiv.org/abs/2303.09556.",
    )
    parser.add_argument(
        "--use_8bit_adam", action="store_true", help="Whether or not to use 8-bit Adam from bitsandbytes."
    )
    parser.add_argument(
        "--allow_tf32",
        action="store_true",
        help=(
            "Whether or not to allow TF32 on Ampere GPUs. Can be used to speed up training. For more information, see"
            " https://pytorch.org/docs/stable/notes/cuda.html#tensorfloat-32-tf32-on-ampere-devices"
        ),
    )
    parser.add_argument("--use_ema", action="store_true", help="Whether to use EMA model.")
    parser.add_argument(
        "--dataloader_num_workers",
        type=int,
        default=0,
        help=(
            "Number of subprocesses to use for data loading. 0 means that the data will be loaded in the main process."
        ),
    )
    parser.add_argument("--adam_beta1", type=float, default=0.9, help="The beta1 parameter for the Adam optimizer.")
    parser.add_argument("--adam_beta2", type=float, default=0.999, help="The beta2 parameter for the Adam optimizer.")
    parser.add_argument("--adam_weight_decay", type=float, default=1e-2, help="Weight decay to use.")
    parser.add_argument("--adam_epsilon", type=float, default=1e-08, help="Epsilon value for the Adam optimizer")
    parser.add_argument("--max_grad_norm", default=1.0, type=float, help="Max gradient norm.")
    parser.add_argument("--push_to_hub", action="store_true", help="Whether or not to push the model to the Hub.")
    parser.add_argument("--hub_token", type=str, default=None, help="The token to use to push to the Model Hub.")
    parser.add_argument(
        "--hub_model_id",
        type=str,
        default=None,
        help="The name of the repository to keep in sync with the local `output_dir`.",
    )
    parser.add_argument(
        "--logging_dir",
        type=str,
        default="logs",
        help=(
            "[TensorBoard](https://www.tensorflow.org/tensorboard) log directory. Will default to"
            " *output_dir/runs/**CURRENT_DATETIME_HOSTNAME***."
        ),
    )
    parser.add_argument(
        "--mixed_precision",
        type=str,
        default=None,
        choices=["no", "fp16", "bf16"],
        help=(
            "Whether to use mixed precision. Choose between fp16 and bf16 (bfloat16). Bf16 requires PyTorch >="
            " 1.10.and an Nvidia Ampere GPU.  Default to the value of accelerate config of the current system or the"
            " flag passed with the `accelerate.launch` command. Use this argument to override the accelerate config."
        ),
    )
    parser.add_argument(
        "--report_to",
        type=str,
        default="tensorboard",
        help=(
            'The integration to report the results and logs to. Supported platforms are `"tensorboard"`'
            ' (default), `"wandb"` and `"comet_ml"`. Use `"all"` to report to all integrations.'
        ),
    )
    parser.add_argument("--local_rank", type=int, default=-1, help="For distributed training: local_rank")
    parser.add_argument(
        "--checkpointing_steps",
        type=int,
        default=500,
        help=(
            "Save a checkpoint of the training state every X updates. These checkpoints are only suitable for resuming"
            " training using `--resume_from_checkpoint`."
        ),
    )
    parser.add_argument(
        "--checkpoints_total_limit",
        type=int,
        default=None,
        help=(
            "Max number of checkpoints to store. Passed as `total_limit` to the `Accelerator` `ProjectConfiguration`."
            " See Accelerator::save_state https://huggingface.co/docs/accelerate/package_reference/accelerator#accelerate.Accelerator.save_state"
            " for more docs"
        ),
    )
    parser.add_argument(
        "--resume_from_checkpoint",
        type=str,
        default=None,
        help=(
            "Whether training should be resumed from a previous checkpoint. Use a path saved by"
            ' `--checkpointing_steps`, or `"latest"` to automatically select the last available checkpoint.'
        ),
    )
    parser.add_argument(
        "--enable_xformers_memory_efficient_attention", action="store_true", help="Whether or not to use xformers."
    )
    parser.add_argument("--noise_offset", type=float, default=0, help="The scale of noise offset.")
    parser.add_argument(
        "--validation_epochs",
        type=int,
        default=1,
        help="Run validation every X epochs.",
    )
    parser.add_argument(
        "--validation_steps",
        type=int,
        default=500,
        help="Run validation every X steps.",
    )
    parser.add_argument(
        "--tracker_project_name",
        type=str,
        default="text2image-fine-tune",
        help=(
            "The `project_name` argument passed to Accelerator.init_trackers for"
            " more information see https://huggingface.co/docs/accelerate/v0.17.0/en/package_reference/accelerator#accelerate.Accelerator"
        ),
    )
    parser.add_argument("--corruption_probability", type=float, default=0.4, help="Probability of a missing pixel.")
    parser.add_argument("--delta_probability", type=float, default=0.1, help="Probability of corrupting an existing pixel.")
    parser.add_argument('--corruption_pattern', type=str, help='Corruption pattern', default='dust')

    args = parser.parse_args()
    args.validation_prompts = args.validation_prompts.split(",")
    env_local_rank = int(os.environ.get("LOCAL_RANK", -1))
    if env_local_rank != -1 and env_local_rank != args.local_rank:
        args.local_rank = env_local_rank

    # Sanity checks
    if args.dataset_name is None and args.train_data_dir is None:
        raise ValueError("Need either a dataset name or a training folder.")

    return args


def main():
    args = parse_args()

    logging_dir = os.path.join(args.output_dir, args.logging_dir)

    accelerator_project_config = ProjectConfiguration(total_limit=args.checkpoints_total_limit)

    accelerator = Accelerator(
        gradient_accumulation_steps=args.gradient_accumulation_steps,
        mixed_precision=args.mixed_precision,
        log_with=args.report_to,
        logging_dir=logging_dir,
        project_config=accelerator_project_config,
    )

    # Make one log on every process with the configuration for debugging.
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        level=logging.INFO,
    )
    logger.info(accelerator.state, main_process_only=False)
    if accelerator.is_local_main_process:
        datasets.utils.logging.set_verbosity_warning()
        transformers.utils.logging.set_verbosity_warning()
        diffusers.utils.logging.set_verbosity_info()
    else:
        datasets.utils.logging.set_verbosity_error()
        transformers.utils.logging.set_verbosity_error()
        diffusers.utils.logging.set_verbosity_error()

    # If passed along, set the training seed now.
    if args.seed is not None:
        set_seed(args.seed)

    # Handle the repository creation
    if accelerator.is_main_process:
        if args.output_dir is not None:
            os.makedirs(args.output_dir, exist_ok=True)

        if args.push_to_hub:
            repo_id = create_repo(
                repo_id=args.hub_model_id or Path(args.output_dir).name, exist_ok=True, token=args.hub_token
            ).repo_id

    # Load scheduler, tokenizer and models.
    if not args.empty_text:
        text_encoder = T5EncoderModel.from_pretrained("DeepFloyd/IF-I-XL-v1.0", subfolder="text_encoder",  device_map="auto", load_in_8bit=True, variant="8bit")
        text_encoder.requires_grad_(False)
        tokenizer = T5Tokenizer.from_pretrained("DeepFloyd/IF-I-XL-v1.0", subfolder="tokenizer")
    else:
        empty_embeds = torch.load("empty_embeds.pt").to(accelerator.device)

    noise_scheduler = DDPMScheduler.from_pretrained("DeepFloyd/IF-I-XL-v1.0", subfolder="scheduler")
    unet = UNet2DConditionModel.from_pretrained("DeepFloyd/IF-I-XL-v1.0", subfolder="unet")
    unet.requires_grad_(True)


    if args.use_ema:
        ema_unet = UNet2DConditionModel.from_pretrained("DeepFloyd/IF-I-XL-v1.0", subfolder="unet")
        ema_unet = EMAModel(ema_unet.parameters(), model_cls=UNet2DConditionModel, model_config=ema_unet.config)

    if args.enable_xformers_memory_efficient_attention:
        if is_xformers_available():
            import xformers

            xformers_version = version.parse(xformers.__version__)
            if xformers_version == version.parse("0.0.16"):
                logger.warn(
                    "xFormers 0.0.16 cannot be used for training in some GPUs. If you observe problems during training, please update xFormers to at least 0.0.17. See https://huggingface.co/docs/diffusers/main/en/optimization/xformers for more details."
                )
            unet.enable_xformers_memory_efficient_attention()
        else:
            raise ValueError("xformers is not available. Make sure it is installed correctly")


    # `accelerate` 0.16.0 will have better support for customized saving
    if version.parse(accelerate.__version__) >= version.parse("0.16.0"):
        # create custom saving & loading hooks so that `accelerator.save_state(...)` serializes in a nice format
        def save_model_hook(models, weights, output_dir):
            if args.use_ema:
                ema_unet.save_pretrained(os.path.join(output_dir, "unet_ema"))

            for i, model in enumerate(models):
                model.save_pretrained(os.path.join(output_dir, "unet"))

                # make sure to pop weight so that corresponding model is not saved again
                weights.pop()

        def load_model_hook(models, input_dir):
            if args.use_ema:
                load_model = EMAModel.from_pretrained(os.path.join(input_dir, "unet_ema"), UNet2DConditionModel)
                ema_unet.load_state_dict(load_model.state_dict())
                ema_unet.to(accelerator.device)
                del load_model

            for i in range(len(models)):
                # pop models so that they are not loaded again
                model = models.pop()

                # load diffusers style into model
                load_model = UNet2DConditionModel.from_pretrained(input_dir, subfolder="unet")
                model.register_to_config(**load_model.config)

                model.load_state_dict(load_model.state_dict())
                del load_model

        accelerator.register_save_state_pre_hook(save_model_hook)
        accelerator.register_load_state_pre_hook(load_model_hook)

    if args.gradient_checkpointing:
        unet.enable_gradient_checkpointing()

    # Enable TF32 for faster training on Ampere GPUs,
    # cf https://pytorch.org/docs/stable/notes/cuda.html#tensorfloat-32-tf32-on-ampere-devices
    if args.allow_tf32:
        torch.backends.cuda.matmul.allow_tf32 = True

    if args.scale_lr:
        args.learning_rate = (
            args.learning_rate * args.gradient_accumulation_steps * args.train_batch_size * accelerator.num_processes
        )

    # Initialize the optimizer
    if args.use_8bit_adam:
        try:
            import bitsandbytes as bnb
        except ImportError:
            raise ImportError(
                "Please install bitsandbytes to use 8-bit Adam. You can do so by running `pip install bitsandbytes`"
            )

        optimizer_cls = bnb.optim.AdamW8bit
    else:
        optimizer_cls = torch.optim.AdamW

    optimizer = optimizer_cls(
        unet.parameters(),
        lr=args.learning_rate,
        betas=(args.adam_beta1, args.adam_beta2),
        weight_decay=args.adam_weight_decay,
        eps=args.adam_epsilon,
    )

    # Get the datasets: you can either provide your own training and evaluation files (see below)
    # or specify a Dataset from the hub (the dataset will be downloaded automatically from the datasets Hub).

    # In distributed training, the load_dataset function guarantees that only one local process can concurrently
    # download the dataset.
    if args.dataset_name is not None:
        # Downloading and loading a dataset from the hub.
        dataset = load_dataset(
            args.dataset_name,
            args.dataset_config_name,
            cache_dir=args.cache_dir,
        )
        # Preprocessing the datasets.
        # We need to tokenize inputs and targets.
        column_names = dataset["train"].column_names

        # 6. Get the column names for input/target.
        dataset_columns = DATASET_NAME_MAPPING.get(args.dataset_name, None)
        if args.image_column is None:
            image_column = dataset_columns[0] if dataset_columns is not None else column_names[0]
        else:
            image_column = args.image_column
            if image_column not in column_names:
                raise ValueError(
                    f"--image_column' value '{args.image_column}' needs to be one of: {', '.join(column_names)}"
                )
        if args.caption_column is None:
            caption_column = dataset_columns[1] if dataset_columns is not None else column_names[1]
        else:
            caption_column = args.caption_column
            if caption_column not in column_names:
                print("Did not find caption column in dataset. This will lead to empty captions.")
                assert args.empty_text == True, "If no caption column is found, empty_text must be set to True."

            # We need to tokenize input captions and transform the images.
            def tokenize_captions(examples, is_train=True):
                captions = []
                for caption in examples[caption_column]:
                    if isinstance(caption, str):
                        captions.append(caption)
                    elif isinstance(caption, (list, np.ndarray)):
                        # take a random caption if there are multiple
                        captions.append(random.choice(caption) if is_train else caption[0])
                    else:
                        captions.append("")

                inputs = tokenizer(
                    captions, max_length=tokenizer.model_max_length, padding="max_length", truncation=True, return_tensors="pt"
                )
                return inputs.input_ids

            # Preprocessing the datasets.
            train_transforms = transforms.Compose(
                [
                    transforms.Resize(args.resolution, interpolation=transforms.InterpolationMode.BILINEAR),
                    transforms.CenterCrop(args.resolution) if args.center_crop else transforms.RandomCrop(args.resolution),
                    transforms.RandomHorizontalFlip() if args.random_flip else transforms.Lambda(lambda x: x),
                    transforms.ToTensor(),
                    transforms.Normalize([0.5], [0.5]),
                ]
            )

            def preprocess_train(examples):
                images = [image.convert("RGB") for image in examples[image_column]]
                examples["pixel_values"] = [train_transforms(image) for image in images]

                if not args.empty_text:
                    examples["input_ids"] = tokenize_captions(examples)
                image_shape = (len(examples["pixel_values"]), 3, 64, 64)
                if args.corruption_pattern == 'dust':
                    corruption_mask = get_random_mask(
                        (len(examples["pixel_values"]), 3, 64, 64), survival_probability=1 - args.corruption_probability,
                        mask_full_rgb=True)
                    extra_mask = get_random_mask((len(examples["pixel_values"]), 3, 64,  64), survival_probability=1 - args.delta_probability,
                                                mask_full_rgb=True)
                elif args.corruption_pattern == 'fixed_box':
                    patch_size = int((args.corruption_probability) * image_shape[-2])
                    corruption_mask = 1 - get_patch_mask(image_shape, patch_size, same_for_all_batch=False)
                    if args.delta_probability > 0:
                        extra_mask = 1 - get_patch_mask(image_shape, patch_size, same_for_all_batch=False)
                    else:
                        extra_mask = corruption_mask

                else:
                    raise NotImplementedError(f"Unknown corruption pattern: {args.corruption_pattern}")
                hat_corruption_mask = extra_mask * corruption_mask
                examples["corruption_masks"] = [mask for mask in corruption_mask]
                examples["hat_corruption_masks"] = [mask for mask in hat_corruption_mask]

                return examples

            with accelerator.main_process_first():
                dataset["train"] = dataset["train"].shuffle(seed=args.seed)
                if args.max_train_samples is not None:
                    dataset["train"] = dataset["train"].select(range(args.max_train_samples))
                # Set the training transforms
                train_dataset = dataset["train"].with_transform(preprocess_train)

            def collate_fn(examples):
                pixel_values = torch.stack([example["pixel_values"] for example in examples])
                pixel_values = pixel_values.to(memory_format=torch.contiguous_format).float()

                if not args.empty_text:
                    input_ids = torch.stack([example["input_ids"] for example in examples])
                else:
                    input_ids = None
                corruption_masks = torch.stack([example["corruption_masks"] for example in examples])
                corruption_masks = corruption_masks.to(memory_format=torch.contiguous_format).float()
                
                hat_corruption_masks = torch.stack([example["hat_corruption_masks"] for example in examples])
                hat_corruption_masks = hat_corruption_masks.to(memory_format=torch.contiguous_format).float()

                return {"pixel_values": pixel_values, "input_ids": input_ids, "corruption_masks": corruption_masks, "hat_corruption_masks": hat_corruption_masks}
    else:
        train_dataset = ImageFolderDataset(path=args.train_data_dir, max_size=args.max_train_samples, random_seed=args.seed, 
                                    corruption_probability=args.corruption_probability, delta_probability=args.delta_probability,
                                    corruption_pattern=args.corruption_pattern, mask_full_rgb=True)
        def collate_fn(examples):
            pixel_values = torch.stack([torch.tensor(example[0]) for example in examples])
            pixel_values = pixel_values.to(memory_format=torch.contiguous_format).float()

            corruption_matrix = torch.stack([torch.tensor(example[2]) for example in examples])
            corruption_masks = corruption_matrix.to(memory_format=torch.contiguous_format).float()

            hat_corruption_matrix = torch.stack([torch.tensor(example[3]) for example in examples])
            hat_corruption_masks = hat_corruption_matrix.to(memory_format=torch.contiguous_format).float()
            return {"pixel_values": pixel_values, "input_ids": None, "corruption_masks": corruption_masks, "hat_corruption_masks": hat_corruption_masks}



    # DataLoaders creation:
    train_dataloader = torch.utils.data.DataLoader(
        train_dataset,
        shuffle=True,
        collate_fn=collate_fn,
        batch_size=args.train_batch_size,
        num_workers=args.dataloader_num_workers,
    )

    # Scheduler and math around the number of training steps.
    overrode_max_train_steps = False
    num_update_steps_per_epoch = math.ceil(len(train_dataloader) / args.gradient_accumulation_steps)
    if args.max_train_steps is None:
        args.max_train_steps = args.num_train_epochs * num_update_steps_per_epoch
        overrode_max_train_steps = True

    lr_scheduler = get_scheduler(
        args.lr_scheduler,
        optimizer=optimizer,
        num_warmup_steps=args.lr_warmup_steps * args.gradient_accumulation_steps,
        num_training_steps=args.max_train_steps * args.gradient_accumulation_steps,
    )

    # Prepare everything with our `accelerator`.
    unet, optimizer, train_dataloader, lr_scheduler = accelerator.prepare(
        unet, optimizer, train_dataloader, lr_scheduler
    )

    if args.use_ema:
        ema_unet.to(accelerator.device)


    # We need to recalculate our total training steps as the size of the training dataloader may have changed.
    num_update_steps_per_epoch = math.ceil(len(train_dataloader) / args.gradient_accumulation_steps)
    if overrode_max_train_steps:
        args.max_train_steps = args.num_train_epochs * num_update_steps_per_epoch
    # Afterwards we recalculate our number of training epochs
    args.num_train_epochs = math.ceil(args.max_train_steps / num_update_steps_per_epoch)

    # We need to initialize the trackers we use, and also store our configuration.
    # The trackers initializes automatically on the main process.
    if accelerator.is_main_process:
        tracker_config = dict(vars(args))
        tracker_config.pop("validation_prompts")
        accelerator.init_trackers(args.tracker_project_name, tracker_config)

    # Train!
    total_batch_size = args.train_batch_size * accelerator.num_processes * args.gradient_accumulation_steps

    logger.info("***** Running training *****")
    logger.info(f"  Num examples = {len(train_dataset)}")
    logger.info(f"  Num Epochs = {args.num_train_epochs}")
    logger.info(f"  Instantaneous batch size per device = {args.train_batch_size}")
    logger.info(f"  Total train batch size (w. parallel, distributed & accumulation) = {total_batch_size}")
    logger.info(f"  Gradient Accumulation steps = {args.gradient_accumulation_steps}")
    logger.info(f"  Total optimization steps = {args.max_train_steps}")
    global_step = 0
    first_epoch = 0

    # Potentially load in the weights and states from a previous save
    if args.resume_from_checkpoint:
        if args.resume_from_checkpoint != "latest":
            path = os.path.basename(args.resume_from_checkpoint)
        else:
            # Get the most recent checkpoint
            dirs = os.listdir(args.output_dir)
            dirs = [d for d in dirs if d.startswith("checkpoint")]
            dirs = sorted(dirs, key=lambda x: int(x.split("-")[1]))
            path = dirs[-1] if len(dirs) > 0 else None

        if path is None:
            accelerator.print(
                f"Checkpoint '{args.resume_from_checkpoint}' does not exist. Starting a new training run."
            )
            args.resume_from_checkpoint = None
        else:
            accelerator.print(f"Resuming from checkpoint {path}")
            accelerator.load_state(os.path.join(args.output_dir, path))
            global_step = int(path.split("-")[1])

            resume_global_step = global_step * args.gradient_accumulation_steps
            first_epoch = global_step // num_update_steps_per_epoch
            resume_step = resume_global_step % (num_update_steps_per_epoch * args.gradient_accumulation_steps)

    # Only show the progress bar once on each machine.
    progress_bar = tqdm(range(global_step, args.max_train_steps), disable=not accelerator.is_local_main_process)
    progress_bar.set_description("Steps")


    for epoch in range(first_epoch, args.num_train_epochs):
        unet.train()
        train_loss = 0.0
        test_loss = 0.0
        unobserved_loss = 0.0
        loss_on_hat_pixels = 0.0
        for step, batch in enumerate(train_dataloader):
            # Skip steps until we reach the resumed step
            if args.resume_from_checkpoint and epoch == first_epoch and step < resume_step:
                if step % args.gradient_accumulation_steps == 0:
                    progress_bar.update(1)
                continue

            with accelerator.accumulate(unet):
                latents = batch["pixel_values"]

                # Sample noise that we'll add to the latents
                noise = torch.randn_like(latents)
                if args.noise_offset:
                    # https://www.crosslabs.org//blog/diffusion-with-offset-noise
                    noise += args.noise_offset * torch.randn(
                        (latents.shape[0], latents.shape[1], 1, 1), device=latents.device
                    )

                bsz = latents.shape[0]
                # Sample a random timestep for each image
                timesteps = torch.randint(0, noise_scheduler.config.num_train_timesteps, (bsz,), device=latents.device)
                timesteps = timesteps.long()

                # Add noise to the latents according to the noise magnitude at each timestep
                # (this is the forward diffusion process)
                
                noisy_latents = noise_scheduler.add_noise(latents, noise, timesteps)
                noisy_latents = batch["hat_corruption_masks"] * noisy_latents

                if not args.empty_text:
                    # Get the text embedding for conditioning
                    encoder_hidden_states = text_encoder(batch["input_ids"])[0]
                else:
                    encoder_hidden_states = empty_embeds.repeat([noisy_latents.shape[0], 1, 1])

                if step == 0:
                    try:
                        os.makedirs(os.path.join(args.output_dir, "artifacts"), exist_ok=True)
                        save_dir = os.path.join(args.output_dir, "artifacts")
                        num_rows = int(np.sqrt(bsz))
                        save_images((batch["pixel_values"]), os.path.join(save_dir, f"dataset.png"), num_rows=num_rows, num_cols=num_rows)
                        save_images((batch["pixel_values"] * batch["corruption_masks"]), 
                                    os.path.join(save_dir, f"dataset_corrupted.png"), num_rows=num_rows, num_cols=num_rows)
                        save_images((batch["pixel_values"] * batch["hat_corruption_masks"]), 
                                    os.path.join(save_dir, f"dataset_hat_corrupted.png"), num_rows=num_rows, num_cols=num_rows)
                    except:
                        print("Can't save images. Is batch size a square number?")

                # Get the target for loss depending on the prediction type
                if noise_scheduler.config.prediction_type == "epsilon":
                    target = noise
                elif noise_scheduler.config.prediction_type == "v_prediction":
                    raise NotImplementedError("Not yet implemented")
                else:
                    raise ValueError(f"Unknown prediction type {noise_scheduler.config.prediction_type}")

                # Predict the noise residual and compute loss
                noise_pred_all = unet(noisy_latents, timesteps, encoder_hidden_states.to(noisy_latents.dtype)).sample

                # get the clean image from the prediction
                alpha_prod_t = noise_scheduler.alphas_cumprod[timesteps.cpu()].reshape(-1, 1, 1, 1).to(accelerator.device)
                beta_prod_t = 1 - alpha_prod_t
                noise_pred, predicted_variance = torch.split(noise_pred_all, noisy_latents.shape[1], dim=1)
                pred_original_sample = batch["hat_corruption_masks"] * (noisy_latents - beta_prod_t ** (0.5) * noise_pred) / alpha_prod_t ** (0.5) + (1 - batch["hat_corruption_masks"]) * noise_pred
                
                if step == 0:
                    exp_name = args.output_dir.split("/")[-1]
                    try:
                        os.makedirs(os.path.join(args.output_dir, "artifacts"), exist_ok=True)
                        save_dir = os.path.join(args.output_dir, "artifacts")
                        save_images(pred_original_sample, 
                                    os.path.join(save_dir, f"{exp_name}_predictions_{global_step}.png"), num_rows=num_rows, num_cols=num_rows)
                        save_images(noisy_latents, 
                                    os.path.join(save_dir, f"{exp_name}_noisy_latents_{global_step}.png"), num_rows=num_rows, num_cols=num_rows)
                        save_images(latents, 
                                    os.path.join(save_dir, f"{exp_name}_latents_{global_step}.png"), num_rows=num_rows, num_cols=num_rows)
                    except:
                        print("Can't save images. Is batch size a square number?")

                loss = F.mse_loss((batch["corruption_masks"] * pred_original_sample).float(), (batch["corruption_masks"] * batch["pixel_values"]).float(), reduction="mean")
                with torch.no_grad():
                    test_loss = F.mse_loss(pred_original_sample.float(), batch["pixel_values"].float(), reduction="mean")
                    unobserved_loss = F.mse_loss((1 - batch["corruption_masks"]) * pred_original_sample, (1 - batch["corruption_masks"]) * batch["pixel_values"], reduction="mean")

                    # get the pixels for which corruption mask is 1 (visible) but hat_corruption_mask is 0 (unobserved)
                    loss_on_hat_pixels = F.mse_loss(( batch["corruption_masks"] * (1 - batch["hat_corruption_masks"]) * pred_original_sample).float(), 
                        (batch["corruption_masks"] * (1 - batch["hat_corruption_masks"]) * batch["pixel_values"]).float(), reduction="mean")

                # Gather the losses across all processes for logging (if we use distributed training).
                avg_loss = accelerator.gather(loss.repeat(args.train_batch_size)).mean()
                avg_test_loss = accelerator.gather(test_loss.repeat(args.train_batch_size)).mean()
                avg_unobserved_loss = accelerator.gather(unobserved_loss.repeat(args.train_batch_size)).mean()
                avg_loss_on_hat_pixels = accelerator.gather(loss_on_hat_pixels.repeat(args.train_batch_size)).mean()
                train_loss += avg_loss.item() / args.gradient_accumulation_steps
                test_loss += avg_test_loss.item() / args.gradient_accumulation_steps
                unobserved_loss = avg_unobserved_loss.item()
                loss_on_hat_pixels = avg_loss_on_hat_pixels.item()
                    
                # Backpropagate
                accelerator.backward(loss)
                if accelerator.sync_gradients:
                    accelerator.clip_grad_norm_(unet.parameters(), args.max_grad_norm)
                optimizer.step()
                lr_scheduler.step()
                optimizer.zero_grad()

            # Checks if the accelerator has performed an optimization step behind the scenes
            if accelerator.sync_gradients:
                if args.use_ema:
                    ema_unet.step(unet.parameters())
                progress_bar.update(1)
                global_step += 1
                accelerator.log({"train_loss": train_loss}, step=global_step)
                accelerator.log({"test_loss": test_loss}, step=global_step)
                accelerator.log({"unobserved_loss": unobserved_loss}, step=global_step)
                accelerator.log({"loss_on_hat_pixels": loss_on_hat_pixels}, step=global_step)

                train_loss = 0.0
                test_loss = 0.0

                if global_step % args.checkpointing_steps == 0:
                    if accelerator.is_main_process:
                        save_path = os.path.join(args.output_dir, f"checkpoint-{global_step}")
                        accelerator.save_state(save_path)
                        logger.info(f"Saved state to {save_path}")

            logs = {"step_loss": loss.detach().item(), "lr": lr_scheduler.get_last_lr()[0]}
            progress_bar.set_postfix(**logs)

            if global_step >= args.max_train_steps:
                break

            if accelerator.is_main_process:
                if args.validation_prompts is not None and (global_step + 1) % args.validation_steps == 0:
                    if args.use_ema:
                        # Store the UNet parameters temporarily and load the EMA parameters to perform inference.
                        ema_unet.store(unet.parameters())
                        ema_unet.copy_to(unet.parameters())
                    print("Saving....")
                    with dnnlib.util.open_url(os.path.join(args.output_dir, f'state_dict_{global_step + 1}.pt'), read_mode='wb') as f:
                        torch.save(accelerator.unwrap_model(unet).state_dict(), f)
                    print("Model saved....")
                    if args.use_ema:
                        # Switch back to the original UNet parameters.
                        ema_unet.restore(unet.parameters())

    accelerator.end_training()


if __name__ == "__main__":
    main()
